# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import logging
import re
from os.path import join

from cmf.monitor import schema
from cmf.monitor.disk_update import DiskUpdate
from cmf.monitor.host.block_device_stats import BlockDeviceStats
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

_BLOCK_DEV_DIR = "/sys/block"
_STAT_FILE = "stat"

class DiskDevice(object):
  """Represents information related to a disk or partition device"""

  def __init__(self, disk_name, partition_name = None, stat_path = None):
    """Creates a DiskDevice for a whole disk or a disk partition

       If partition_name is provided, the DiskDevice models a single partition
       rather than the entire disk.

       param stat_path is provided for testing purposes

       Constructor will throw an exception if an error occurs.
    """
    self._device = partition_name or disk_name
    self._device_stats_prev = None
    self._device_stats_curr = BlockDeviceStats()

    if stat_path is None:
      # The stat file is in:
      # /sys/<disk>/stat              for disks      (disk=sda)
      # /sys/<disk>/<partition>/stat  for partitions (disk=sda, partition=sda3)
      if partition_name is None:
        self._stat_path = join(_BLOCK_DEV_DIR, disk_name, _STAT_FILE)
      else:
        self._stat_path = join(_BLOCK_DEV_DIR, disk_name, partition_name, _STAT_FILE)
    else:
      self._stat_path = stat_path

    if self.collect_update() is None:
      raise Exception('Unable to create DiskDevice ' + self._device)

  def device(self):
    return self._device

  def collect_update(self):
    try:
      stat_file = open(self._stat_path)
      try:
        stats = [long(s.strip()) for s in stat_file.readline().split()]
        # This will be caught and logged below if the assert fails
        assert len(stats) == 11, "Unexpected stat file contents '%s'" % (self._stat_path)
        self._device_stats_prev = self._device_stats_curr
        self._device_stats_curr = BlockDeviceStats(stats)

        (await, await_r, await_w) = self.calculate_await_times(
          self._device_stats_curr,
          self._device_stats_prev)

        service_time = self.calculate_service_time(self._device_stats_curr,
                                                   self._device_stats_prev)

        update = DiskUpdate(self._device)
        update.add_metric(schema.AWAIT_TIME, await)
        update.add_metric(schema.AWAIT_READ_TIME, await_r)
        update.add_metric(schema.AWAIT_WRITE_TIME, await_w)
        update.add_metric(schema.SERVICE_TIME, service_time)
        update.add_metric(schema.READ_BYTES,
                          self._device_stats_curr.get_read_bytes())
        update.add_metric(schema.WRITE_BYTES,
                          self._device_stats_curr.get_write_bytes())
        update.add_metric(schema.READ_IOS,
                          self._device_stats_curr.get_read_ios())
        update.add_metric(schema.WRITE_IOS,
                          self._device_stats_curr.get_write_ios())
        return update
      finally:
        stat_file.close()
    except Exception:
      THROTTLED_LOG.exception("Failed to collect disk stats for device '%s'", self._device)
      return None

  @classmethod
  def calculate_await_times(cls, curr_stats, prev_stats):
    """Returns a tuple containing the read and write await times"""

    await_r_time = 0
    new_read_ios = curr_stats.get_read_ios() - prev_stats.get_read_ios()
    new_read_ticks = curr_stats.get_read_ticks() - prev_stats.get_read_ticks()
    if new_read_ios > 0:
      await_r_time = new_read_ticks / float(new_read_ios)

    await_w_time = 0
    new_write_ios = curr_stats.get_write_ios() - prev_stats.get_write_ios()
    new_write_ticks = curr_stats.get_write_ticks() - prev_stats.get_write_ticks()
    if new_write_ios > 0:
      await_w_time = new_write_ticks / float(new_write_ios)

    await_time = 0
    total_ios = new_read_ios + new_write_ios
    total_ticks = new_read_ticks + new_write_ticks
    if total_ios > 0:
      await_time = total_ticks / float(total_ios)

    return (await_time, await_r_time, await_w_time)

  @classmethod
  def calculate_service_time(cls, curr_stats, prev_stats):
    new_ios = curr_stats.get_nr_ios() - prev_stats.get_nr_ios()
    if new_ios > 0:
      delta_io_ticks = curr_stats.get_io_ticks() - \
                       prev_stats.get_io_ticks()
      return delta_io_ticks / float(new_ios)
    else:
      return None
